import hashlib
import textwrap
from typing import Any

from fastapi.responses import PlainTextResponse
from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, model_validator


def get_environment(*args, **kwargs):
    return Environment(
        *args,
        **kwargs,
        variable_start_string="'_{{",
        variable_end_string="}}_'",
        comment_start_string="{#=%",
        comment_end_string="%=#}",
    )


class BashFunction(BaseModel):
    body: str
    name: str | None = None
    context: dict | None = None

    @model_validator(mode="before")
    @classmethod
    def update_name(cls, data: Any) -> Any:
        if not data.get("name"):
            data["name"] = "f" + hashlib.md5(data["body"].encode()).hexdigest()
        return data


def function(func: dict, indent: int = 0, spaces: int = 4) -> str:
    func = BashFunction(**func)
    text = func.body
    if func.context:
        text = get_environment().from_string(func.body).render(func.context)
    template = f"\n{func.name}() {{"
    template += textwrap.indent(textwrap.dedent(text), " " * spaces)
    template += "}"
    return textwrap.indent(template, " " * spaces * indent)


def echo(text: str, indent: int = 0, spaces: int = 4) -> str:
    lines = textwrap.dedent(text).splitlines()
    lines = [textwrap.indent(f'echo "{line}"', " " * spaces * indent) for line in lines]
    return "\n" + "\n".join(lines)


class TemplateEngine:
    def __init__(self, dir_path: str) -> None:
        self.dir_path = dir_path
        self.env = get_environment(loader=FileSystemLoader(dir_path))

    def render(self, template_name: str, context: dict) -> str:
        template = self.env.get_template(template_name)
        template.globals.update(
            {
                "function": function,
                "echo": echo,
            }
        )
        return PlainTextResponse(template.render(context))



import abc
import os
import textwrap
from typing import List

from pydantic import BaseModel, field_validator

from mqtt_rest.plugins.template_engine import BashFunction, TemplateEngine

THIS_DIR = os.path.dirname(__file__)

helper_templates = TemplateEngine(
    os.path.join(
        THIS_DIR,
        "_helper_commands",
    ),
)


class Command(BaseModel):
    command: str
    install_func: BashFunction | None = None


class Installer(BaseModel):
    url: str
    dependencies: List[Command] | None = None
    description: str | None = None

    def render(self):
        return helper_templates.render("install.sh", self.model_dump())

    @field_validator("description", mode="before")
    def update_description(cls, value):
        if value:
            return textwrap.dedent(value)
        return value


class BaseJob(BaseModel, abc.ABC):
    url: str
    job_func: BashFunction
    freq2cron_func: BashFunction
    need_root: bool = False

    @abc.abstractmethod
    def render(self):
        pass


class SingleJob(BaseJob):
    def render(self):
        return helper_templates.render("single_job.sh", self.model_dump())


class MultiJob(BaseJob):
    job_options_func: BashFunction

    def render(self):
        return helper_templates.render("multi_job.sh", self.model_dump())


def get_cron_frequency(freq: str):
    options = []
    if "m" in freq:
        options.append('m) echo "*/$number * * * *";;')
    if "h" in freq:
        options.append('h) echo "0 */$number * * *";;')
    if "d" in freq:
        options.append('d) echo "0 0 */$number * *";;')
    return BashFunction(
        body="""
        local frequency=$1
        local unit=${frequency: -1}
        local number=${frequency:0:${#frequency}-1}

        case $unit in
            {% for option in options -%}
            '_{{ option }}_'
            {% endfor -%}
            *) echo "Invalid frequency unit. Please use one of '_{{valids}}_'."
            exit 1 ;;
        esac
        """,
        name="get_cron_frequency",
        context={"options": options, "valids": ",".join(freq)},
    )


JOB_FREQUENCY_MINUTE_HOUR_DAY = get_cron_frequency("mhd")
JOB_FREQUENCY_HOUR_DAY = get_cron_frequency("hd")
JOB_FREQUENCY_DAY = get_cron_frequency("d")



"""
Parse the output of the command
and return a dictionary of the sensor values
"""

import re
from typing import Dict, Tuple


def parse_group(text: str) -> Tuple[str, float]:
    device = data_units_written = None
    for line in text.splitlines():
        if ":" not in line:
            continue
        key, value = line.split(":", 1)
        if key.strip() == "device":
            device = value.strip()
        elif key.strip() == "total bytes written":
            data_units_written = round(float(re.search(r"(\d+)", value).group(1)) / (1024**3), 3)
    return device, data_units_written


def parse_report(output: str) -> Dict[str, float]:
    sensors = {}
    for group in re.split(r"\n-+\n", output):
        name, value = parse_group(group.lower())
        if name and value:
            sensors[f"{name}"] = value
    return sensors
